@@ -55,7 +55,7 @@ int pam_modutil_drop_priv(pam_handle_t *pamh, struct _ykpam_privs *privs, struct
     privs->saved_egid = getegid();
 
     if ((privs->saved_euid == pw->pw_uid) && (privs->saved_egid == pw->pw_gid)) {
-        D (privs->debug_file, "Privilges already dropped, pretend it is all right");
+        D (privs->debug_file, "Privileges already dropped, pretend it is all right");
         return 0;
     }
 
@@ -66,7 +66,7 @@ int pam_modutil_drop_priv(pam_handle_t *pamh, struct _ykpam_privs *privs, struct
     }
 
     if (privs->saved_groups_length > SAVED_GROUPS_MAX_LEN) {
-        D (privs->debug_file, "to many groups, limiting.");
+        D (privs->debug_file, "too many groups, limiting.");
         privs->saved_groups_length = SAVED_GROUPS_MAX_LEN;
     }
 
