@@ -1,5 +1,5 @@
 /* Written by Simon Josefsson <simon@yubico.com>.
- * Copyright (c) 2006-2016 Yubico AB
+ * Copyright (c) 2006-2019 Yubico AB
  * Copyright (c) 2011 Tollef Fog Heen <tfheen@err.no>
  * All rights reserved.
  *
@@ -28,6 +28,8 @@
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
+#define _GNU_SOURCE
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
@@ -43,6 +45,7 @@
 
 #include "util.h"
 #include "drop_privs.h"
+#include "ykbzero.h"
 
 #include <ykclient.h>
 
@@ -51,6 +54,7 @@
 #include <yubikey.h>
 /* for yubikey pbkdf2*/
 #include <ykpbkdf2.h>
+#include <ykpers-version.h>
 #endif /* HAVE_CR */
 
 /* Libtool defines PIC for shared objects */
@@ -108,7 +112,10 @@ struct cfg
   int verbose_otp;
   int try_first_pass;
   int use_first_pass;
+  int always_prompt;
   int nullok;
+  int ldap_starttls;
+  int ldap_bind_as_user;
   const char *auth_file;
   const char *capath;
   const char *cainfo;
@@ -122,9 +129,17 @@ struct cfg
   const char *ldap_filter;
   const char *ldap_cacertfile;
   const char *ldapdn;
+  const char *ldap_clientcertfile;
+  const char *ldap_clientkeyfile;
   const char *user_attr;
   const char *yubi_attr;
   const char *yubi_attr_prefix;
+  const char *mysql_server;
+  int mysql_port;
+  const char *mysql_user;
+  const char *mysql_password;
+  const char *mysql_database;
+
   unsigned int token_id_length;
   enum key_mode mode;
   const char *chalresp_path;
@@ -136,6 +151,13 @@ struct cfg
 #endif
 #define DBG(x...) if (cfg->debug) { D(cfg->debug_file, x); }
 
+/* Helper to free memory used by pam_set_data */
+static void
+setcred_free (pam_handle_t *pamh, void *ptr, int err)
+{
+  free (ptr);
+}
+
 /*
  * Authorize authenticated OTP_ID for login as USERNAME using AUTHFILE.
  *
@@ -148,8 +170,19 @@ authorize_user_token (struct cfg *cfg,
 		      pam_handle_t *pamh)
 {
   int retval = AUTH_ERROR;
-
-  if (cfg->auth_file)
+  if (cfg->mysql_server)
+    {
+#ifdef HAVE_MYSQL
+      /* Administrator had configured the database and specified is name
+        as an argument for this module.
+      */
+      DBG ("Using Mariadb or Mysql Database");
+      retval = check_user_token_mysql(cfg->mysql_server, cfg->mysql_port, cfg->mysql_user, cfg->mysql_password, cfg->mysql_database, username, otp_id, cfg->debug, cfg->debug_file);
+#else
+      DBG (("Trying to use MYSQL, but this function is not compiled in pam_yubico!!"));
+#endif
+    }
+  else if (cfg->auth_file)
     {
       /* Administrator had configured the file and specified is name
          as an argument for this module.
@@ -165,6 +198,7 @@ authorize_user_token (struct cfg *cfg,
       size_t buflen = sizeof(buf);
       int pwres;
       PAM_MODUTIL_DEF_PRIVS(privs);
+      struct stat st;
 
       pwres = getpwnam_r (username, &pass, buf, buflen, &p);
       if (p == NULL) {
@@ -180,7 +214,7 @@ authorize_user_token (struct cfg *cfg,
          ..... i.e. ~/.yubico/authorized_yubikeys
        */
       if (! get_user_cfgfile_path (NULL, "authorized_yubikeys", p, &userfile)) {
-	DBG ("Failed figuring out per-user cfgfile");
+	DBG ("Failed to figure out per-user cfgfile");
 	return AUTH_ERROR;
       }
 
@@ -190,7 +224,11 @@ authorize_user_token (struct cfg *cfg,
         goto free_out;
       }
 
-      retval = check_user_token (userfile, username, otp_id, cfg->debug, cfg->debug_file);
+      if (lstat (userfile, &st) != 0 && errno == ENOENT) {
+        retval = AUTH_NO_TOKENS;
+      } else {
+        retval = check_user_token (userfile, username, otp_id, cfg->debug, cfg->debug_file);
+      }
 
       if(pam_modutil_regain_priv(pamh, &privs)) {
         DBG ("could not restore privileges");
@@ -228,7 +266,8 @@ free_out:
 static int
 authorize_user_token_ldap (struct cfg *cfg,
 			   const char *user,
-			   const char *token_id)
+			   const char *token_id,
+			   pam_handle_t *pamh)
 {
   int retval = AUTH_ERROR;
 #ifdef HAVE_LIBLDAP
@@ -244,6 +283,7 @@ authorize_user_token_ldap (struct cfg *cfg,
   struct berval **vals;
   int rc;
   size_t i;
+  int j;
 
   char *filter = NULL;
   char *find = NULL;
@@ -283,22 +323,30 @@ authorize_user_token_ldap (struct cfg *cfg,
   ldap_set_option (ld, LDAP_OPT_PROTOCOL_VERSION, &protocol);
 
   if (cfg->ldap_uri && cfg->ldap_cacertfile) {
-    /* Set CA CERTFILE.  This makes ldaps work when using ldap_uri */
+    /* Set CA CERTFILE. This makes ldaps work when using ldap_uri */
     ldap_set_option (0, LDAP_OPT_X_TLS_CACERTFILE, cfg->ldap_cacertfile);
   }
-  /* Bind anonymously to the LDAP server. */
-  if (cfg->ldap_bind_user && cfg->ldap_bind_password) {
-    DBG ("try bind with: %s:[%s]", cfg->ldap_bind_user, cfg->ldap_bind_password);
-    rc = ldap_simple_bind_s (ld, cfg->ldap_bind_user, cfg->ldap_bind_password);
-  } else {
-    DBG ("try bind anonymous");
-    rc = ldap_simple_bind_s (ld, NULL, NULL);
-  }
-  if (rc != LDAP_SUCCESS)
-    {
-      DBG ("ldap_simple_bind_s: %s", ldap_err2string (rc));
+
+  if (cfg->ldap_clientcertfile && cfg->ldap_clientkeyfile) {
+    rc = ldap_set_option (NULL, LDAP_OPT_X_TLS_CERTFILE, cfg->ldap_clientcertfile);
+    if (rc != LDAP_SUCCESS) {
+      DBG ("tls_certfile: %s", ldap_err2string (rc));
       goto done;
     }
+    rc = ldap_set_option (NULL, LDAP_OPT_X_TLS_KEYFILE, cfg->ldap_clientkeyfile);
+    if (rc != LDAP_SUCCESS) {
+      DBG ("tls_keyfile: %s", ldap_err2string (rc));
+      goto done;
+    }
+  }
+
+  if (cfg->ldap_starttls) {
+    rc = ldap_start_tls_s (ld, NULL, NULL);
+    if (rc != LDAP_SUCCESS) {
+      DBG ("ldap_start_tls: %s", ldap_err2string (rc));
+      goto done;
+    }
+  }
 
   /* Allocation of memory for search strings depending on input size */
   if (cfg->user_attr && cfg->yubi_attr && cfg->ldapdn) {
@@ -307,11 +355,41 @@ authorize_user_token_ldap (struct cfg *cfg,
       DBG ("Failed allocating %zu bytes", i);
       goto done;
     }
-    sprintf (find, "%s=%s,%s", cfg->user_attr, user, cfg->ldapdn);
+    j = snprintf (find, i, "%s=%s,%s", cfg->user_attr, user, cfg->ldapdn);
+    if (j < 0 || j >= i) {
+      DBG ("Failed to format string");
+      goto done;
+    }
     filter = NULL;
   } else if (cfg->ldapdn) {
     find = strdup(cfg->ldapdn); /* allow free later */
   }
+
+  /* Bind to the LDAP server. */
+  if (cfg->ldap_bind_as_user && cfg->user_attr && cfg->yubi_attr && cfg->ldapdn) {
+    /* Bind as the user logging in with their password they provided to PAM */
+    const char *bind_password = NULL;
+    rc = pam_get_item (pamh, PAM_AUTHTOK, (const void **) &bind_password);
+    if (rc != PAM_SUCCESS) {
+      DBG ("pam_get_item failed to retrieve password: %s", pam_strerror (pamh, rc));
+      goto done;
+    }
+    DBG ("try bind as user with: %s", find);
+    rc = ldap_simple_bind_s (ld, find, bind_password);
+  } else if (cfg->ldap_bind_user && cfg->ldap_bind_password) {
+    /* Bind with a provided username and password */
+    DBG ("try bind with: %s:[%s]", cfg->ldap_bind_user, cfg->ldap_bind_password);
+    rc = ldap_simple_bind_s (ld, cfg->ldap_bind_user, cfg->ldap_bind_password);
+  } else {
+    DBG ("try anonymous bind");
+    rc = ldap_simple_bind_s (ld, NULL, NULL);
+  }
+  if (rc != LDAP_SUCCESS)
+    {
+      DBG ("ldap_simple_bind_s: %s", ldap_err2string (rc));
+      goto done;
+    }
+
   if (cfg->ldap_filter) {
     filter = filter_printf(cfg->ldap_filter, user);
     scope = LDAP_SCOPE_SUBTREE;
@@ -375,9 +453,9 @@ authorize_user_token_ldap (struct cfg *cfg,
 		        {
 		          retval = AUTH_NOT_FOUND;
 		        }
-		      if(token_id && !strncmp (token_id, vals[i]->bv_val + yubi_attr_prefix_len, strlen (vals[i]->bv_val + yubi_attr_prefix_len)))
+		      if(token_id && !strncmp(token_id, vals[i]->bv_val + yubi_attr_prefix_len, cfg->token_id_length))
 		        {
-		          DBG ("Token Found :: %s", vals[i]->bv_val);
+		          DBG ("Token found :: %s", vals[i]->bv_val);
 		          retval = AUTH_FOUND;
 		        }
 		    }
@@ -432,7 +510,7 @@ display_error(pam_handle_t *pamh, const char *message, struct cfg *cfg) {
   msg[0].msg = (char *) message; /* on some systems, pam_message.msg isn't const */
   msg[0].msg_style = PAM_ERROR_MSG;
   retval = conv->conv(1, pmsg, &resp, conv->appdata_ptr);
-  
+
   if (retval != PAM_SUCCESS) {
     DBG("conv returned error: %s", pam_strerror (pamh, retval));
     return retval;
@@ -511,7 +589,7 @@ do_challenge_response(pam_handle_t *pamh, struct cfg *cfg, const char *username)
   }
 
   if (! init_yubikey(&yk)) {
-    DBG("Failed initializing YubiKey");
+    DBG("Failed to initialize YubiKey");
     goto out;
   }
 
@@ -521,7 +599,7 @@ do_challenge_response(pam_handle_t *pamh, struct cfg *cfg, const char *username)
   }
 
   if (! get_user_challenge_file (yk, cfg->chalresp_path, p, &userfile, cfg->debug_file)) {
-    DBG("Failed getting user challenge file for user %s", username);
+    DBG("Failed to get user challenge file for user %s", username);
     goto out;
   }
 
@@ -530,7 +608,7 @@ do_challenge_response(pam_handle_t *pamh, struct cfg *cfg, const char *username)
   /* Drop privileges before opening user file (if we're not using system-wide dir). */
   if (!cfg->chalresp_path) {
     if (pam_modutil_drop_priv(pamh, &privs, p)) {
-      DBG ("could not drop privileges");
+      DBG ("Could not drop privileges");
       goto out;
     }
   }
@@ -571,7 +649,7 @@ do_challenge_response(pam_handle_t *pamh, struct cfg *cfg, const char *username)
 
   if (!cfg->chalresp_path) {
     if (pam_modutil_regain_priv(pamh, &privs)) {
-      DBG ("could not restore privileges");
+      DBG ("Could not restore privileges");
       goto out;
     }
   }
@@ -579,7 +657,7 @@ do_challenge_response(pam_handle_t *pamh, struct cfg *cfg, const char *username)
   if (! challenge_response(yk, state.slot, state.challenge, state.challenge_len,
 			   true, true, false,
 			   buf, sizeof(buf), &response_len)) {
-    DBG("Challenge-response FAILED");
+    DBG("Challenge-response failed");
     goto out;
   }
 
@@ -597,7 +675,7 @@ do_challenge_response(pam_handle_t *pamh, struct cfg *cfg, const char *username)
   if (memcmp(buf, state.response, state.response_len) == 0) {
     ret = PAM_SUCCESS;
   } else {
-    DBG("Unexpected C/R response : %s", response_hex);
+    DBG("Unexpected response: %s", response_hex);
     goto out;
   }
 
@@ -605,15 +683,15 @@ do_challenge_response(pam_handle_t *pamh, struct cfg *cfg, const char *username)
 
   errstr = "Error generating new challenge, please check syslog or contact your system administrator";
   if (generate_random(state.challenge, sizeof(state.challenge))) {
-    DBG("Failed generating new challenge!");
+    DBG("Failed to generate new challenge!");
     goto out;
   }
 
-  errstr = "Error communicating with Yubikey, please check syslog or contact your system administrator";
+  errstr = "Error communicating with YubiKey, please check syslog or contact your system administrator";
   if (! challenge_response(yk, state.slot, state.challenge, CR_CHALLENGE_SIZE,
 			   true, true, false,
 			   buf, sizeof(buf), &response_len)) {
-    DBG("Second challenge-response FAILED");
+    DBG("Second challenge-response failed");
     goto out;
   }
 
@@ -630,6 +708,7 @@ do_challenge_response(pam_handle_t *pamh, struct cfg *cfg, const char *username)
   /*
    * Write the challenge and response we will expect the next time to the state file.
    */
+  errstr = "Error updating YubiKey challenge, please check syslog or contact your system administrator";
   if (response_len > sizeof(state.response)) {
     DBG("Got too long response ??? (%u/%zu)", response_len, sizeof(state.response));
     goto out;
@@ -642,7 +721,7 @@ do_challenge_response(pam_handle_t *pamh, struct cfg *cfg, const char *username)
   /* Drop privileges before creating new challenge file. */
   if (!cfg->chalresp_path) {
     if (pam_modutil_drop_priv(pamh, &privs, p)) {
-        DBG ("could not drop privileges");
+        DBG ("Could not drop privileges");
         goto out;
     }
   }
@@ -661,11 +740,11 @@ do_challenge_response(pam_handle_t *pamh, struct cfg *cfg, const char *username)
   }
 
   if (fchmod (fd, st.st_mode) != 0) {
-      DBG ("could not set correct file permissions");
+      DBG ("Could not set correct file permissions");
       goto restpriv_out;
   }
   if (fchown (fd, st.st_uid, st.st_gid) != 0) {
-      DBG ("could not set correct file ownership");
+      DBG ("Could not set correct file ownership");
       goto restpriv_out;
   }
 
@@ -675,7 +754,6 @@ do_challenge_response(pam_handle_t *pamh, struct cfg *cfg, const char *username)
     goto restpriv_out;
   }
 
-  errstr = "Error updating Yubikey challenge, please check syslog or contact your system administrator";
   if (! write_chalresp_state (f, &state))
     goto out;
   if (fclose(f) < 0) {
@@ -695,7 +773,7 @@ do_challenge_response(pam_handle_t *pamh, struct cfg *cfg, const char *username)
 restpriv_out:
   if (!cfg->chalresp_path) {
     if (pam_modutil_regain_priv(pamh, &privs)) {
-        DBG (("could not restore privileges"));
+        DBG (("Could not restore privileges"));
     }
   }
 
@@ -705,8 +783,8 @@ restpriv_out:
       syslog(LOG_ERR, "USB error: %s", yk_usb_strerror());
       DBG("USB error: %s", yk_usb_strerror());
     } else {
-      syslog(LOG_ERR, "Yubikey core error: %s", yk_strerror(yk_errno));
-      DBG("Yubikey core error: %s", yk_strerror(yk_errno));
+      syslog(LOG_ERR, "YubiKey core error: %s", yk_strerror(yk_errno));
+      DBG("YubiKey core error: %s", yk_strerror(yk_errno));
     }
   }
 
@@ -714,8 +792,8 @@ restpriv_out:
     display_error(pamh, errstr, cfg);
 
   if (errno) {
-    syslog(LOG_ERR, "Challenge response failed: %s", strerror(errno));
-    DBG("Challenge response failed: %s", strerror(errno));
+    syslog(LOG_ERR, "Challenge-response failed: %s", strerror(errno));
+    DBG("Challenge-response failed: %s", strerror(errno));
   }
 
   if (yk)
@@ -734,6 +812,9 @@ restpriv_out:
 static void
 parse_cfg (int flags, int argc, const char **argv, struct cfg *cfg)
 {
+  struct stat st;
+  FILE *file = NULL;
+  int fd = -1;
   int i;
 
   memset (cfg, 0, sizeof(struct cfg));
@@ -758,8 +839,14 @@ parse_cfg (int flags, int argc, const char **argv, struct cfg *cfg)
 	cfg->try_first_pass = 1;
       if (strcmp (argv[i], "use_first_pass") == 0)
 	cfg->use_first_pass = 1;
+      if (strcmp (argv[i], "always_prompt") == 0)
+	cfg->always_prompt = 1;
       if (strcmp (argv[i], "nullok") == 0)
 	cfg->nullok = 1;
+      if (strcmp (argv[i], "ldap_starttls") == 0)
+	cfg->ldap_starttls = 1;
+      if (strcmp (argv[i], "ldap_bind_as_user") == 0)
+	cfg->ldap_bind_as_user = 1;
       if (strncmp (argv[i], "authfile=", 9) == 0)
 	cfg->auth_file = argv[i] + 9;
       if (strncmp (argv[i], "capath=", 7) == 0)
@@ -784,6 +871,10 @@ parse_cfg (int flags, int argc, const char **argv, struct cfg *cfg)
 	cfg->ldap_filter = argv[i] + 12;
       if (strncmp (argv[i], "ldap_cacertfile=", 16) == 0)
         cfg->ldap_cacertfile = argv[i] + 16;
+      if (strncmp (argv[i], "ldap_clientcertfile=", 20) == 0)
+        cfg->ldap_clientcertfile = argv[i] + 20;
+      if (strncmp (argv[i], "ldap_clientkeyfile=", 19) == 0)
+        cfg->ldap_clientkeyfile = argv[i] + 19;
       if (strncmp (argv[i], "ldapdn=", 7) == 0)
 	cfg->ldapdn = argv[i] + 7;
       if (strncmp (argv[i], "user_attr=", 10) == 0)
@@ -800,6 +891,17 @@ parse_cfg (int flags, int argc, const char **argv, struct cfg *cfg)
 	cfg->mode = CLIENT;
       if (strncmp (argv[i], "chalresp_path=", 14) == 0)
 	cfg->chalresp_path = argv[i] + 14;
+      if (strncmp (argv[i], "mysql_server=", 13) == 0)
+	cfg->mysql_server = argv[i] + 13;
+      if (strncmp (argv[i], "mysql_port=", 11) == 0)
+	sscanf (argv[i], "mysql_port=%u", &cfg->mysql_port);
+      if (strncmp (argv[i], "mysql_user=", 11) == 0)
+	cfg->mysql_user = argv[i] + 11;
+      if (strncmp (argv[i], "mysql_password=", 15) == 0)
+	cfg->mysql_password = argv[i] + 15;
+      if (strncmp (argv[i], "mysql_database=", 15) == 0)
+	cfg->mysql_database = argv[i] + 15;
+
       if (strncmp (argv[i], "debug_file=", 11) == 0)
         {
           const char *filename = argv[i] + 11;
@@ -813,22 +915,15 @@ parse_cfg (int flags, int argc, const char **argv, struct cfg *cfg)
             }
           else
             {
-              struct stat st;
-              int fd;
-              FILE *file;
-              if(lstat(filename, &st) == 0)
+              fd = open(filename, O_WRONLY | O_APPEND | O_CLOEXEC | O_NOFOLLOW | O_NOCTTY);
+              if (fd >= 0 && (fstat(fd, &st) == 0) && S_ISREG(st.st_mode))
                 {
-                  if(S_ISREG(st.st_mode))
+                  file = fdopen(fd, "a");
+                  if(file != NULL)
                     {
-                      fd = open(filename, O_WRONLY | O_CREAT | O_APPEND | O_CLOEXEC, S_IRUSR | S_IWUSR | S_IRGRP);
-                      if (fd >= 0)
-                        {
-                          file = fdopen(fd, "a");
-                          if (file)
-                            {
-                              cfg->debug_file = file;
-                            }
-                        }
+                      cfg->debug_file = file;
+                      file = NULL;
+                      fd = -1;
                     }
                 }
             }
@@ -847,7 +942,10 @@ parse_cfg (int flags, int argc, const char **argv, struct cfg *cfg)
   DBG ("verbose_otp=%d", cfg->verbose_otp);
   DBG ("try_first_pass=%d", cfg->try_first_pass);
   DBG ("use_first_pass=%d", cfg->use_first_pass);
+  DBG ("always_prompt=%d", cfg->always_prompt);
   DBG ("nullok=%d", cfg->nullok);
+  DBG ("ldap_starttls=%d", cfg->ldap_starttls);
+  DBG ("ldap_bind_as_user=%d", cfg->ldap_bind_as_user);
   DBG ("authfile=%s", cfg->auth_file ? cfg->auth_file : "(null)");
   DBG ("ldapserver=%s", cfg->ldapserver ? cfg->ldapserver : "(null)");
   DBG ("ldap_uri=%s", cfg->ldap_uri ? cfg->ldap_uri : "(null)");
@@ -856,6 +954,8 @@ parse_cfg (int flags, int argc, const char **argv, struct cfg *cfg)
   DBG ("ldap_filter=%s", cfg->ldap_filter ? cfg->ldap_filter : "(null)");
   DBG ("ldap_cacertfile=%s", cfg->ldap_cacertfile ? cfg->ldap_cacertfile : "(null)");
   DBG ("ldapdn=%s", cfg->ldapdn ? cfg->ldapdn : "(null)");
+  DBG ("ldap_clientcertfile=%s", cfg->ldap_clientcertfile ? cfg->ldap_clientcertfile : "(null)");
+  DBG ("ldap_clientkeyfile=%s", cfg->ldap_clientkeyfile ? cfg->ldap_clientkeyfile : "(null)");
   DBG ("user_attr=%s", cfg->user_attr ? cfg->user_attr : "(null)");
   DBG ("yubi_attr=%s", cfg->yubi_attr ? cfg->yubi_attr : "(null)");
   DBG ("yubi_attr_prefix=%s", cfg->yubi_attr_prefix ? cfg->yubi_attr_prefix : "(null)");
@@ -867,6 +967,16 @@ parse_cfg (int flags, int argc, const char **argv, struct cfg *cfg)
   DBG ("token_id_length=%u", cfg->token_id_length);
   DBG ("mode=%s", cfg->mode == CLIENT ? "client" : "chresp" );
   DBG ("chalresp_path=%s", cfg->chalresp_path ? cfg->chalresp_path : "(null)");
+  DBG ("mysql_server=%s", cfg->mysql_server ? cfg->mysql_server : "(null)");
+  DBG ("mysql_port=%d", cfg->mysql_port);
+  DBG ("mysql_user=%s", cfg->mysql_user ? cfg->mysql_user : "(null)");
+  DBG ("mysql_database=%s", cfg->mysql_database ? cfg->mysql_database : "(null)");
+
+  if (fd != -1)
+    close(fd);
+
+  if (file != NULL)
+    fclose(file);
 }
 
 PAM_EXTERN int
@@ -900,7 +1010,7 @@ pam_sm_authenticate (pam_handle_t * pamh,
 
   if (cfg->token_id_length > MAX_TOKEN_ID_LEN)
   {
-    DBG ("configuration error: token_id_length too long. Maximum acceptable value : %u", MAX_TOKEN_ID_LEN);
+    DBG ("Configuration error: token_id_length too long. Maximum acceptable value : %u", MAX_TOKEN_ID_LEN);
     retval = PAM_AUTHINFO_UNAVAIL;
     goto done;
   }
@@ -915,9 +1025,10 @@ pam_sm_authenticate (pam_handle_t * pamh,
 
   if (cfg->mode == CHRESP) {
 #if HAVE_CR
+    DBG ("libykpers version: %s", ykpers_check_version(NULL));
     retval = do_challenge_response(pamh, cfg, user);
 #else
-    DBG ("no support for challenge/response");
+    DBG ("no support for challenge-response");
     retval = PAM_AUTH_ERR;
 #endif
     goto done;
@@ -932,7 +1043,7 @@ pam_sm_authenticate (pam_handle_t * pamh,
 	      pam_strerror (pamh, retval));
 	  goto done;
 	}
-      DBG ("get password returned: %s", password);
+      DBG ("get password returned: /* not logged */");
     }
 
   if (cfg->use_first_pass && password == NULL)
@@ -944,7 +1055,7 @@ pam_sm_authenticate (pam_handle_t * pamh,
 
   if(ykclient_global_init() != YKCLIENT_OK)
     {
-      DBG ("Failed initializing ykclient library");
+      DBG ("Failed to initlaize ykclient library");
       retval = PAM_AUTHINFO_UNAVAIL;
       goto done;
     }
@@ -1018,34 +1129,36 @@ pam_sm_authenticate (pam_handle_t * pamh,
   /* check if the user has at least one associated token id */
   /* we set otp_id to NULL so that no matches will ever be found
    * but AUTH_NO_TOKENS will be returned if there are no tokens for the user */
-  if (cfg->ldapserver != NULL || cfg->ldap_uri != NULL)
-    valid_token = authorize_user_token_ldap (cfg, user, NULL);
-  else
-    valid_token = authorize_user_token (cfg, user, NULL, pamh);
+  if (!cfg->always_prompt) {
+    if (cfg->ldapserver != NULL || cfg->ldap_uri != NULL)
+      valid_token = authorize_user_token_ldap (cfg, user, NULL, pamh);
+    else
+      valid_token = authorize_user_token (cfg, user, NULL, pamh);
 
-  switch(valid_token)
-    {
-    case AUTH_ERROR:
-      DBG ("Internal error while looking for user tokens");
-      retval = PAM_AUTHINFO_UNAVAIL;
-      goto done;
-    case AUTH_NOT_FOUND:
-      /* User has associated tokens, so continue */
-      DBG ("Tokens found for user");
-      break;
-    case AUTH_NO_TOKENS:
-      DBG ("No tokens found for user");
-      if (cfg->nullok) {
-        retval = PAM_IGNORE;
-      } else {
-        retval = PAM_USER_UNKNOWN;
+    switch(valid_token)
+      {
+      case AUTH_ERROR:
+        DBG ("Internal error while looking for user tokens");
+        retval = PAM_AUTHINFO_UNAVAIL;
+        goto done;
+      case AUTH_NOT_FOUND:
+        /* User has associated tokens, so continue */
+        DBG ("Tokens found for user");
+        break;
+      case AUTH_NO_TOKENS:
+        DBG ("No tokens found for user");
+        if (cfg->nullok) {
+          retval = PAM_IGNORE;
+        } else {
+          retval = PAM_USER_UNKNOWN;
+        }
+        goto done;
+      default:
+        DBG ("Unhandled value while looking for user tokens");
+        retval = PAM_AUTHINFO_UNAVAIL;
+        goto done;
       }
-      goto done;
-    default:
-      DBG ("Unhandled value while looking for user tokens");
-      retval = PAM_AUTHINFO_UNAVAIL;
-      goto done;
-    }
+  }
 
   if (password == NULL)
     {
@@ -1100,7 +1213,7 @@ pam_sm_authenticate (pam_handle_t * pamh,
 
   password_len = strlen (password);
 
-  /* In case the input was systempassword+YubiKeyOTP, we want to skip over
+  /* In case the input was systempassword+OTP, we want to skip over
      "systempassword" when copying the token_id and OTP to separate buffers */
   if(password_len > cfg->token_id_length + TOKEN_OTP_LEN)
     {
@@ -1115,8 +1228,6 @@ pam_sm_authenticate (pam_handle_t * pamh,
   /* Copy only public ID into otp_id. Destination buffer is zeroed. */
   strncpy (otp_id, password + skip_bytes, cfg->token_id_length);
 
-  DBG ("OTP: %s ID: %s ", otp, otp_id);
-
   /* user entered their system password followed by generated OTP? */
   if (password_len > TOKEN_OTP_LEN + cfg->token_id_length)
     {
@@ -1144,17 +1255,18 @@ pam_sm_authenticate (pam_handle_t * pamh,
 
   /* authorize the user with supplied token id */
   if (cfg->ldapserver != NULL || cfg->ldap_uri != NULL)
-    valid_token = authorize_user_token_ldap (cfg, user, otp_id);
+    valid_token = authorize_user_token_ldap (cfg, user, otp_id, pamh);
   else
     valid_token = authorize_user_token (cfg, user, otp_id, pamh);
 
   switch(valid_token)
     {
     case AUTH_FOUND:
+      DBG ("OTP: %s ID: %s ", otp, otp_id);
       DBG ("Token is associated to the user. Validating the OTP...");
       rc = ykclient_request (ykc, otp);
       DBG ("ykclient return value (%d): %s", rc, ykclient_strerror (rc));
-      DBG ("ykclient url used: %s", ykclient_get_last_url(ykc));
+      DBG ("ykclient URL used: %s", ykclient_get_last_url(ykc));
 
       switch (rc)
       {
@@ -1196,7 +1308,12 @@ pam_sm_authenticate (pam_handle_t * pamh,
 
 done:
   if (onlypasswd)
-    free(onlypasswd);
+    {
+      insecure_memzero(onlypasswd, strlen(onlypasswd));
+      free(onlypasswd);
+    }
+  insecure_memzero(otp, sizeof(otp));
+  insecure_memzero(otp_id, sizeof(otp_id));
   if (templates > 0)
     {
       size_t i;
@@ -1218,7 +1335,21 @@ done:
       retval = PAM_SUCCESS;
     }
   DBG ("done. [%s]", pam_strerror (pamh, retval));
-  pam_set_data (pamh, "yubico_setcred_return", (void*)(intptr_t)retval, NULL);
+
+  int* pretval = malloc (sizeof(int));
+  if (pretval)
+    {
+      *pretval = retval;
+      if (pam_set_data (pamh, "yubico_setcred_return", (void*)pretval,
+	   setcred_free) != PAM_SUCCESS)
+        {
+          DBG ("pam_set_data failed setting setcred_return: %d", retval);
+        }
+    }
+  else
+    {
+	DBG ("Failed allocating memory for setcred_return status");
+    }
 
   if (resp)
     {
@@ -1253,11 +1384,12 @@ pam_sm_acct_mgmt(pam_handle_t *pamh, int flags, int argc, const char **argv)
 {
   struct cfg cfg_st;
   struct cfg *cfg = &cfg_st; /* for DBG macro */
-  int retval;
-  int rc = pam_get_data(pamh, "yubico_setcred_return", (const void**)&retval);
+  int retval = PAM_AUTH_ERR;
+  const void *pretval = NULL;
+  int rc = pam_get_data (pamh, "yubico_setcred_return", &pretval);
 
   parse_cfg (flags, argc, argv, cfg);
-  if (rc == PAM_SUCCESS && retval == PAM_SUCCESS) {
+  if (rc == PAM_SUCCESS && pretval && *(const int *)pretval == PAM_SUCCESS) {
     DBG ("pam_sm_acct_mgmt returning PAM_SUCCESS");
     retval = PAM_SUCCESS;
   } else {
